/*
 * spruked - mach-spruked.c
 *
 * Machine-dependend code for spruked board
 *
 * Copyright (C) 2013  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2013  Karol Budniak <karol@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/spruked.h>

#include "spruked.h"

/*
 * Bandwidth saving
 *
 * The UKED outputs may change very fast, so we need high sampling
 * frequency. However, there can be huge periods without changes when
 * everything works correctly. So we use simple bandwidth saving,
 * after SAVER_ACTIVE milliseconds we change sampling period to
 * SAVER_PERIOD.
 *
 * In bandwidth saving mode we count from SAVER_ACTIVE to SAVER_MAX.
 *
 * The minimal frequency allowed by AR-3c is 50 Hz - 20 ms period.
 *
 * Those values are in milliseconds.
 */
#define SAVER_ACTIVE	40
#define SAVER_PERIOD	20
#define SAVER_MAX	(SAVER_ACTIVE + SAVER_PERIOD)

#define PORT_LO		P0
#define PORT_HI		P2

static unsigned char addr;

static u8 spruked_get_addr(void)
{
	unsigned char address = 0x00;

	P3 |= 0xE4;
	address = (P3 & 0xE0) >> 4;
	address = address | ((P3 & 0x04) >> 2);
	return ~address;
}

struct input_state {
	u8 hi;
	u8 lo;
	u8 counter;
};

static struct input_state old_state;

void spruked_mach_poll(void)
{
	struct spruked_skb __kernel *skb;
	u8 __kernel *p;
	u8 p0;
	u8 p1;

	p0 = PORT_LO;
	p1 = PORT_HI;
	old_state.counter += spruked_bus_cycle;
	if (old_state.lo != p0 || old_state.hi != p1) {
		old_state.counter = 0;
	} else {
		/*
		 * We transmit at data at nominal speed up to SAVER_ACTIVE
		 * miliseconds.
		 */
		if (old_state.counter >= SAVER_ACTIVE) {
			if (old_state.counter < SAVER_MAX)
				return;
			old_state.counter = SAVER_ACTIVE;
		}
	}

	skb = spruked_alloc_skb();
	if (!skb) {
		old_state.counter = 0;
		return;
	}

	p = skb->payload;
	*p++ = addr << 4;

	*p++ = p0;
	*p++ = p1;

	old_state.lo = p0;
	old_state.hi = p1;
	spruked_xmit(skb);
}

void spruked_mach_link_state(unsigned char ok)
{
	P1_2 = (ok != 0);
}

static const unsigned int spd_table[] = {
	B115200, B115200, B115200, B115200
};
static const unsigned char bus_cycle_table[] = {
	1, 2, 5, 10
};

/*
 * initialize spruked board - we do it at subsystem level because
 * we want to change serial speed as early as possible and we need
 * to do it after full low-level architecture initialization.
 */
static void spruked_init(void)
{
	unsigned char q = 0x00;

	P1 |= 0x18;
	q = (~P1 >> 3) & 3;
	ser_set_spd(spd_table[q]);
	spruked_bus_cycle = bus_cycle_table[q];

	addr = spruked_get_addr();
}
subsys_initcall(spruked_init)
