/*
 * tmos - i51.c
 *
 * SLIP for i51 serial
 * 
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/port.h>
#include <tmos/crc16.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <asm/task.h>
#include <tmos/errno.h>
#include <tmos/net.h>
#include <asm/irqs.h>
#include <limits.h>

#ifndef dir_tx
#define dir_tx()	do { } while (0)
#endif
#ifndef dir_rx
#define dir_rx()	do { } while (0)
#endif

/* SLIP protocol characters. */
#define END             0300		/* indicates end of frame	*/
#define ESC             0333		/* indicates byte stuffing	*/
#define ESC_END         0334		/* ESC ESC_END means END 'data'	*/
#define ESC_ESC         0335		/* ESC ESC_ESC means ESC 'data'	*/

#define SLIP_MTU		CONFIG_SLIP_MTU

#if CONFIG_SLIP_CRC16
#define SLIP_MINTU		3
#else
#define SLIP_MINTU		1
#endif

#define TXP_END 0xff

#ifdef CONFIG_KERNEL_DATA
static uint8_t __idata msg_buffer[SLIP_MTU];
#else
static uint8_t __kernel msg_buffer[SLIP_MTU];
#endif
static uint8_t __kernel msg_size = 0;
static uint8_t __kernel msg_txp = TXP_END;	/* transmit pointer */
static __bit rx_lock = 0;
static __bit esc = 0;

#ifdef CONFIG_SLIP_ADDR_FILTER

#ifdef CONFIG_SLIP_ADDR_FILTER_RUNTIME
static uint8_t slip_addr = CONFIG_SLIP_ADDR_FILTER_ADDR;
#define slip_address()	slip_addr
#else
#define slip_address()	CONFIG_SLIP_ADDR_FILTER_ADDR
#endif

#ifdef CONFIG_SLIP_ADDR_FILTER_BROADCAST
#define msg_filter() \
	((msg_buffer[0] == slip_address()) || !msg_buffer[0])
#else
#define msg_filter() (msg_buffer[0] == slip_address())
#endif
#else
#define msg_filter() (1)
#endif

/**
 * receive_char - receive character from UART
 */
static void receive_char(void) __using(I51_SERIAL_BANK)
{
	unsigned char tmp;

	tmp = SBUF;

	/*
	 * This implementation uses only one shared buffer. We must
	 * skip this character when buffer is locked.
	 */
	if (rx_lock)
		return;

	/*
	 * check for end of frame
	 */
	if (tmp == END) {
		/*
		 * skip frames with invalid size or frames to other nodes.
		 */
		if (((unsigned char)(msg_size - SLIP_MINTU) <= SLIP_MTU
					- SLIP_MINTU)
				&& msg_filter()) {
			/*
			 * possible good frame, lock buffer and deffer
			 * to poll()
			 */
			rx_lock = 1;
			msg_txp = TXP_END;
		} else {
			msg_size = 0;
		}
		esc = 0;
		return;
	}

	if (esc) {
		esc = 0;
		/*
		 * decode ESCAPE sequence
		 */
		if (tmp == ESC_ESC) {
			tmp = ESC;
		} else if (tmp == ESC_END) {
			tmp = END;
		}
	} else {
		/* check for ESCAPE sequence */
		if (tmp == ESC) {
			esc = 1;
			return;
		}
	}

	if (msg_size < SLIP_MTU) {
		msg_buffer[msg_size] = tmp;
		msg_size++;
	} else {
		msg_size = SLIP_MTU + 1;
	}
}

/**
 * transmit_char - transmit character
 *
 * This function reads data from transmit FIFO, performs escaping
 * of reserved characters and sends it to UART.
 *
 * This function is called from high priority interrupt context.
 */
static void transmit_char(void) __using(I51_SERIAL_BANK)
{
	unsigned char tmp;

	/*
	 * send payload bytes - escaping of reserved END and ESC
	 * characters must be performed.
	 */
	if (msg_txp < msg_size) {
		tmp = msg_buffer[msg_txp];

		if (esc) {
			/* send second byte of escape sequence */
			esc = 0;
			msg_txp++;
			if (tmp == ESC) {
				__tx_send_raw(ESC_ESC);
				return;
			} else {
				__tx_send_raw(ESC_END);
				return;
			}
		}

		/* start escape sequence in case of reserved character */
		if (tmp == END || tmp == ESC) {
			esc = 1;
			__tx_send_raw(ESC);
			return;
		}

		msg_txp++;
		__tx_send_raw(tmp);
		return;
	}

	/*
	 * after payload send one END character and disable transmitter.
	 * Currently disabling transmitter can be up to 1 bit too early.
	 * In case of RS-232 or properly biased RS-485 this should not be
	 * a problem.
	 *
	 * FIXME: don't cut stop bit.
	 */
	if (msg_txp == msg_size) {
		__tx_send_raw(END);
		msg_txp++;
	} else {
		dir_rx();
		msg_size = 0;
		rx_lock = 0;
	}
}

/**
 * seq_irq - serial interrupt
 * 
 * This interrupt is usually called in high priority context.
 */
void ser_irq(void) __interrupt(SERIAL_INTERRUPT) __using(I51_SERIAL_BANK)
{
	if (RI) {
		RI = 0;
		receive_char();
	}

	if (TI) {
		TI = 0;
		transmit_char();
	}
}

char slip_i51_ioctl(unsigned char cmd, void __kernel *arg) __reentrant
{
	char ret = -EINVAL;

	switch (cmd) {
#ifdef CONFIG_SLIP_ADDR_FILTER_RUNTIME
	case NET_IOCTL_SET_ADDR_BYTE:
		slip_addr = *((uint8_t __kernel *) arg);
		ret = 0;
		break;
#endif
	case NET_IOCTL_SET_SPEED:
		ret = 0;
		switch (*((uint8_t __kernel *) arg)) {
#ifdef B9600
		case NET_SPEED_9600:
			ser_set_spd(B9600);
			break;
#endif
#ifdef B19200
		case NET_SPEED_19200:
			ser_set_spd(B19200);
			break;
#endif
#ifdef B38400
		case NET_SPEED_38400:
			ser_set_spd(B38400);
			break;
#endif
#ifdef B57600
		case NET_SPEED_57600:
			ser_set_spd(B57600);
			break;
#endif
#ifdef B115200
		case NET_SPEED_115200:
			ser_set_spd(B115200);
			break;
#endif
#ifdef B125000
		case NET_SPEED_125000:
			ser_set_spd(B125000);
			break;
#endif
#ifdef B230400
		case NET_SPEED_230400:
			ser_set_spd(B230400);
			break;
#endif
#ifdef B250000
		case NET_SPEED_250000:
			ser_set_spd(B250000);
			break;
#endif
		default:
			ret = -EINVAL;
		}
		break;
	}
	return ret;
}

void slip_i51_poll(void) __reentrant
{
	if (rx_lock && (msg_txp == TXP_END)) {
		net_size_t ret = 0;
		uint16_t __kernel *crc;
		struct net_msg msg;

		/* setup message */
		msg_init(&msg);		
		msg_set_interface(&msg, NET_IFACE_SERIAL0);
		msg_set_buffer(&msg, (void __kernel *)msg_buffer);
		msg_set_head(&msg, 0);

#if CONFIG_SLIP_CRC16
		msg_set_tail(&msg, msg_size - 2);
		msg_set_size(&msg, SLIP_MTU - 2);
		crc = (uint16_t __kernel *)(msg_buffer + msg_size - 2);
		if (le16_to_cpu(*crc) == crc16_kernel(0xffff,
					(void __kernel *)msg_buffer,
					msg_size - 2)) {
			ret = net_xmit(NET_PROTO_MODBUS, &msg);
			if (ret) {
				msg_size = ret + 2;
				crc = (uint16_t __kernel *)(msg_buffer + ret);
				*crc = cpu_to_le16(crc16_kernel(0xffff,
							(void __kernel *)
							msg_buffer, ret));
			}
		}
#else
		msg_set_tail(&msg, msg_size);
		msg_set_size(&msg, SLIP_MTU);
		ret = net_xmit(NET_PROTO_MODBUS, &msg);
		if (ret)
			msg_size = ret;
#endif

		i51_serial_interrupt_disable();
		if (ret) {
			dir_tx();
			msg_txp = 0;

			tx_send_raw(END);
		} else {
			msg_size = 0;
			rx_lock = 0;
		}
		i51_serial_interrupt_enable();
	}
}

static void slip_i51_init(void) __reentrant
{
	dir_rx();
	i51_serial_interrupt_enable();
}

device_initcall(slip_i51_init)
