/*
 * tmos - i51.c
 *
 * spruked protocol for i51 serial
 *
 * Copyright (C) 2013  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <tmos/spruked.h>
#include <tmos/net.h>
#include <tmos/errno.h>
#include <tmos/sched.h>
#include <asm/irqs.h>
	
#if I51_SERIAL_BANK == HIGH_PRIORITY_BANK
#define spruked_irq_disable()	do { } while (0)
#define spruked_irq_enable()	do { } while (0)
#else
#define spruked_irq_disable()	local_irq_disable()
#define spruked_irq_enable()	local_irq_enable()
#endif

/*
 * The spruked protocol uses constant size frames - 4 bytes each
 */
#define SPRUKED_FRAME_LEN	4

static u8 __kernel *rx_frame;
static u8 __kernel *tx_frame;

static __bit tx_busy;
__bit spruked_tx_pending;
static __bit synchronized;

#ifndef SPRUKED_SYNC
#define SPRUKED_SYNC		1
#endif

#ifndef align_kptr
#define align_kptr(p, x) ((void __kernel *)((unsigned char)(p) \
			& (unsigned char)(~((x) - 1))))
#endif

void ser_irq(void) __serial_irq_attr __naked
{
	struct spruked_skb __kernel *skb;
	u8 __kernel *t;
	unsigned char tmp;

	irq_enter(I51_SERIAL_BANK);

	if (TI) {
		TI = 0;
		if (!spruked_tx_pending) {
			tx_busy = 0;
		} else {
			SBUF = *tx_frame;
			tx_frame++;
			tmp = (unsigned char) tx_frame;
			if ((tmp & (SPRUKED_FRAME_LEN - 1)) == 0) {
				skb = (struct spruked_skb __kernel *) tx_frame;
				skb--;
				__spruked_consume_skb(skb);
				smp_wmb();
				spruked_tx_pending = 0;
				spruked_tx_completed();
			}
		}
	}

	if (RI) {
		RI = 0;
		tmp = SBUF;
		t = rx_frame;
		if (tmp & SPRUKED_SYNC) {
			t = align_kptr(t, 4);
			synchronized = 1;
		}
		if (synchronized) {
			*t = tmp;
			t++;
			if (!((unsigned char)t & (SPRUKED_FRAME_LEN - 1))) {
				/*
				 * The spruked_rx_ready should be after setting
				 * spruked_rx_skb, but we are on UP. We use
				 * smb_*() barrier only to distinguish between
				 * read and write barriers, but we expect
				 * that both are just compiler barriers.
				 */
				t -= SPRUKED_FRAME_LEN;
				if (!spruked_rx_skb) {
					__spruked_alloc_skb(skb);
					if (skb) {
						spruked_rx_skb =
							(struct spruked_skb *) t;
						spruked_received_frame();
						t = skb->payload;
					}
				}
				synchronized = 0;
			}
		}
		rx_frame = t;
	}

	irq_exit(I51_SERIAL_BANK);
}

void spruked_xmit_start(struct spruked_skb __kernel *skb) __using(SOFTIRQ_BANK)
{
	cpu_flags_t flags;

	local_irq_save(flags);
	spruked_tx_pending = 1;
	tx_frame = skb->payload;
	if (!tx_busy) {
		tx_busy = 1;
		SBUF = skb->payload[0];
		tx_frame++;
	}
	local_irq_restore(flags);
}

char spruked_i51_ioctl(unsigned char cmd, void __kernel *arg) __reentrant
{
	char ret = -EINVAL;

	switch (cmd) {
	case NET_IOCTL_SET_SPEED:
		ret = 0;
		switch (*((uint8_t __kernel *) arg)) {
		case NET_SPEED_115200:
			ser_set_spd(B115200);
		default:
			ret = -EINVAL;
		}
		break;
	}
	return ret;
}

static void spruked_i51_init(void)
{
	rx_frame = spruked_alloc_skb()->payload;
	i51_serial_interrupt_enable();
}
device_initcall(spruked_i51_init)
