/*
 * tmos - i51.c
 *
 * Copyright (C) 2005..2012  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/port.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <tmos/sched.h>
#include <tmos/errno.h>
#include <asm/delay.h>
#include <limits.h>

#ifdef CONFIG_SERIAL_TIMER3
#include <asm/aduc841.h>
#endif

#define XON	0x11
#define XOFF	0x13

#if CONFIG_SERIAL_PARITY_EVEN

#define tx_parity() \
	do { \
		__asm__(mov c, _P); \
		__asm__(mov _TB8, c); \
	} while (0)

#elif CONFIG_SERIAL_PARITY_ODD

#define tx_parity() \
	do { \
		__asm__(mov c, _P); \
		__asm__(cpl c); \
		__asm__(mov _TB8, c); \
	} while (0)

#elif CONFIG_SERIAL_PARITY_SPACE
#define tx_parity() do { TB8 = 0; } while (0)
#elif CONFIG_SERIAL_PARITY_MARK
#define tx_parity() do { TB8 = 1; } while (0)
#else
#define tx_parity() do { } while (0)
#endif

#ifdef CONFIG_SERIAL_I51_STREAM

#if CONFIG_SERIAL_I51_PAGE_BUFFERS
#define RX_RING_SIZE	(PAGE_SIZE)
#define RX_RING_MASK	((unsigned char)((unsigned int) RX_RING_SIZE - 1))
#define TX_RING_SIZE	(PAGE_SIZE)
#define TX_RING_MASK	((unsigned char)((unsigned int) TX_RING_SIZE - 1))

static uint8_t __xdata *rx_ring;
static uint8_t __xdata *tx_ring;
#else

#if CONFIG_KERNEL_DATA
# define SER_BUFS_DATA		1
#elif CONFIG_KERNEL_XDATA
# define SER_BUFS_XDATA		1
#endif

#define RX_RING_SIZE	(1 << CONFIG_SERIAL_I51_RX_BUFFER_SHIFT)
#define RX_RING_MASK	(RX_RING_SIZE - 1)
#define TX_RING_SIZE	(1 << CONFIG_SERIAL_I51_TX_BUFFER_SHIFT)
#define TX_RING_MASK	(TX_RING_SIZE - 1)

static uint8_t __kernel rx_ring[RX_RING_SIZE];
static uint8_t __kernel tx_ring[TX_RING_SIZE];
#endif

uint8_t __kernel rx_head;
uint8_t __kernel rx_tail;
uint8_t __kernel tx_head;
uint8_t __kernel tx_tail;

#if CONFIG_SERIAL_I51_POLL
#define serial_wait() \
	do { \
		ser_irq(); \
		schedule(); \
	} while (0)
#else
#define serial_wait()	schedule()
#endif

#ifdef CONFIG_SERIAL_I51_FLOW_SW

#if RX_RING_SIZE <= 32
#define RX_RING_MIN_FREE	RX_RING_SIZE
#elif RX_RING_SIZE <= 64
#define RX_RING_MIN_FREE	(RX_RING_SIZE - 4)
#else
#define RX_RING_MIN_FREE	(RX_RING_SIZE - 16)
#endif

static u8 ser_fc;	/* serial flow control */

/**
 * ser_rx_wait - wait for character to receive
 *
 * This function should be called in kernel mode with interrupts enabled.
 */
void ser_rx_wait(void) __preempt
{
	unsigned int c = 1000;

	for (;;) {
		serial_wait();
		if (rx_head != rx_tail)
			break;
		if (!c) {
			if (!ser_fc)
				ser_fc = XON;
			c = 10000;
		}
		udelay(10);
		c--;
	}
}

void ser_xoff(void) __reentrant
{
	unsigned char timeout;
	unsigned char tail;

	/* send XOFF character */
	ser_fc = XOFF;
	ser_irq();

retry:
	timeout = 250;
	tail = rx_tail;
	do {
		serial_wait();
		udelay(10);
		if (rx_tail != tail)
			goto retry;
		timeout--;
	} while (timeout > 0);
}

#else

/**
 * ser_rx_wait - wait for character to receive
 *
 * This function should be called in kernel mode with interrupts enabled.
 */
#define ser_rx_wait() \
	do { \
		while (rx_head == rx_tail) \
			serial_wait(); \
	} while (0)

#endif

/**
 * __ser_rxc - receive single character
 *
 * There MUST be a character in receive buffer. This function is not
 * reentrant - should be called in kernel mode with preemption disabled.
 */
static char __ser_rxc(void) __preempt
{
	char ret;

	ret = rx_ring[rx_head];
	rx_head = (rx_head + 1) & RX_RING_MASK;
	return ret;
}

#if CONFIG_SERIAL_DEPRECATED
/**
 * ser_rxc - receive single character
 */
char ser_rxc(void) __preempt
{
	char ret;

	ser_rx_wait();
	ret = __ser_rxc();
	return ret;
}
#endif

#define IMM #

#if __INLINE_ASM && SER_BUFS_DATA
#define tx_send() \
	do { \
		__asm__(mov	a, IMM _tx_ring); \
		__asm__(add	a, _tx_head); \
		__asm__(mov	r0, a); \
		__asm__(mov	a, @r0); \
		tx_parity(); \
		__asm__(mov	SBUF, a); \
	} while (0)
#elif __INLINE_ASM && SER_BUFS_XDATA
#define tx_send() \
	do { \
		__asm__(mov	dptr, IMM _tx_head); \
		__asm__(movx	a, @dptr); \
		__asm__(add	a, IMM _tx_ring); \
		__asm__(mov	dpl, a); \
		__asm__(clr	a); \
		__asm__(addc	a, IMM (_tx_ring >> 8)); \
		__asm__(mov	dph, a); \
		__asm__(movx	a, @dptr); \
		tx_parity(); \
		__asm__(mov SBUF, a); \
	} while (0);
#else
#define tx_send() do { SBUF = tx_ring[tx_head]; } while (0)
#endif

static __bit transmiter_ready = 1;

/*
 * ser_txc - transmit character
 * @c: character
 *
 * This function should be called from kernel context with interrupts
 * enabled.
 */
void ser_txc(char c) __preempt
{
	uint8_t tmp;

	for (;;) {
		preempt_disable();
		tmp = (tx_tail + 1) & TX_RING_MASK;
		if (tmp != tx_head)
			break;
		preempt_enable();
		serial_wait();
	}
	tx_ring[tx_tail] = c;
	smp_wmb();
	tx_tail = tmp;

	if (transmiter_ready) {
		transmiter_ready = 0;
		smp_wmb();
		tx_send();
	}
	preempt_enable();
}

#ifdef CONFIG_SERIAL_DATA
char ser_rxd(uint8_t __data *buf, unsigned char len) __reentrant
{
	char __data *t;
	unsigned char tmp;

	t = (char __data *) buf;
	ser_rx_wait();
	for (tmp = 0; tmp < len && ser_rx_poll(); tmp++, t++)
		*t = __ser_rxc();
	return tmp;
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rxd_t(uint8_t __data *buf, unsigned char len, jiffies_t jiffies) REENTRANT
{
	char __data *t;
	unsigned char tmp;
	jiffies_t t0;

	t = (char __data *) buf;
	for (tmp = 0; tmp < len; tmp++, t++) {
		t0 = get_jiffies();
		while (!ser_rx_poll()) {
			if (get_jiffies() - t0 > jiffies)
				return tmp;
			serial_wait();
		}
		*t = __ser_rxc();
	}
	return tmp;
}
#endif

void ser_txd(const uint8_t __data *buf, char len) __reentrant
{
	while (len--)
		ser_txc(*buf++);
}
#endif

#ifdef CONFIG_SERIAL_XDATA
char ser_rxxd(uint8_t __xdata *buf, unsigned char len) __reentrant
{
	char __xdata *t;
	unsigned char tmp;

	t = (char __xdata *) buf;
	ser_rx_wait();
	for (tmp = 0; tmp < len && ser_rx_poll(); tmp++, t++)
		*t = __ser_rxc();
	return tmp;
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rxxd_t(uint8_t __xdata *buf, unsigned char len, jiffies_t jiffies) __reentrant
{
	char __xdata *t;
	unsigned char tmp;
	jiffies_t t0;

	t = (char __xdata *) buf;
	for (tmp = 0; tmp < len; tmp++, t++) {
		t0 = get_jiffies();
		while (!ser_rx_poll()) {
			if (get_jiffies() - t0 > jiffies)
				return tmp;
			serial_wait();
		}
		*t = __ser_rxc();
	}
	return tmp;
}
#endif

void ser_txxd(const uint8_t __xdata *buf, char len) __reentrant
{
	while (len--)
		ser_txc(*buf++);
}
#endif

#ifdef CONFIG_SERIAL_GENERIC
char ser_rx(uint8_t *buf, unsigned char len) __reentrant
{
	char *t;
	unsigned char tmp;

	t = (char *) buf;
	ser_rx_wait();
	for (tmp = 0; tmp < len && ser_rx_poll(); tmp++, t++)
		*t = __ser_rxc();
	return tmp;
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rx_t(uint8_t *buf, unsigned char len, jiffies_t jiffies) __reentrant
{
	char *t;
	unsigned char tmp;
	jiffies_t t0;

	t = (char *) buf;
	for (tmp = 0; tmp < len; tmp++, t++) {
		t0 = get_jiffies();
		while (!ser_rx_poll()) {
			if (get_jiffies() - t0 > jiffies)
				return tmp;
			serial_wait();
		}
		*t = __ser_rxc();
	}
	return tmp;
}
#endif
#endif

#if SERIAL_WANT_SER_TX
void __ser_tx(const uint8_t *buf, char len) __reentrant
{
	while (len--)
		ser_txc(*buf++);
}
#endif

#if CONFIG_SERIAL_DEPRECATED
void ser_txs(const char *str) __preempt
{
	while (*str != '\0')
		ser_txc(*str++);
}
#endif

void ser_irq(void) __serial_irq_attr
{
	unsigned char tmp;

	if (RI) {
		RI = 0;
		tmp = rx_tail;
		tmp++;
		tmp &= RX_RING_MASK;
		if (tmp != rx_head) {
			rx_ring[rx_tail] = SBUF;
			rx_tail = tmp;
		} else {
			tmp = SBUF;
		}

#if CONFIG_SERIAL_I51_FLOW_SW
		if (((unsigned char)(rx_head - rx_tail - 1) & RX_RING_MASK)
				< RX_RING_MIN_FREE)
			ser_fc = XOFF;
#endif
	}

#if CONFIG_SERIAL_I51_FLOW_SW
	if (TI) {
		TI = 0;

		if (ser_fc) {
			__tx_send_raw(ser_fc);
			ser_fc = 0;
		} else if (tx_head != tx_tail) {
			tx_head++;
			tx_head &= TX_RING_MASK;
			if (tx_head != tx_tail)
				__tx_send_raw(tx_ring[tx_head]);
			else
				transmiter_ready = 1;
		} else {
			transmiter_ready = 1;
		}
	} else {
		if (transmiter_ready && ser_fc) {
			transmiter_ready = 0;
			__tx_send_raw(ser_fc);
			ser_fc = 0;
		}
	}
#else
	if (TI) {
		TI = 0;

		tx_head++;
		tx_head &= TX_RING_MASK;
		if (tx_head != tx_tail)
			tx_send();
		else
			transmiter_ready = 1;
	}
#endif

	set_need_resched();
}

void ser_flush(void) __preempt
{
	while (!transmiter_ready)
		serial_wait();
}
#else

void ser_txc(char c) __preempt
{
	TI = 0;
	SBUF = c;
	while (!TI)
		cpu_relax();
}

#if CONFIG_SERIAL_DEPRECATED
char ser_rxc(void) __reentrant
{
	char ret;

	while (!RI)
		cpu_relax();
	ret = SBUF;
	RI = 0;
	return ret;
}
#endif

#ifdef CONFIG_SERIAL_SEND_RAW
void tx_send_raw(unsigned char x) __reentrant
{
	DPL = x;
	__asm__(mov a, _DPL);
	tx_parity();
	__asm__(mov SBUF, a);
}
#endif

#endif

#ifdef CONFIG_SERIAL_SEND_RAW
void __tx_send_raw(unsigned char x) __using(I51_SERIAL_BANK) __reentrant
{
	DPL = x;
	__asm__(mov a, _DPL);
	tx_parity();
	__asm__(mov SBUF, a);
}
#endif

void ser_set_spd(uint16_t spd) __reentrant
{
#if defined(CONFIG_SERIAL_TIMER3)
	T3CON = spd >> 8;
	T3FD = spd & 0xff;
#elif defined(CONFIG_USE_BRG)
	BRL = spd;
#elif defined(CONFIG_SERIAL_TIMER2)
	TL2 = RCAP2L = spd & 0xff;
	TH2 = RCAP2H = spd >> 8;
#else
	TH1 = spd;
#endif
}

#if CONFIG_SERIAL_I51_PAGE_BUFFERS
static void serial_allocate_buffers(void)
{
	page_t page;

	page = page_alloc();
	BUG_ON(page == PAGE_NULL);
	rx_ring = page_to_virt(page);

	page = page_alloc();
	BUG_ON(page == PAGE_NULL);
	tx_ring = page_to_virt(page);
}
#else
#define serial_allocate_buffers() do { } while (0)
#endif

static void i51_serial_init(void)
{
	serial_allocate_buffers();

	PCON |= SMOD;

#if defined(CONFIG_SERIAL_PARITY_NONE)
	SCON = 0x50;
#else
	SCON = 0xd0;
#endif

#if defined(CONFIG_SERIAL_TIMER3)
#elif defined(CONFIG_USE_BRG)
	BDRCON |= BDRCON_BRR | BDRCON_TBCK | BDRCON_RBCK | BDRCON_SPD
		| BDRCON_SRC;
#elif defined(CONFIG_SERIAL_TIMER2)
	T2CON = 0x30;
#else
	TMOD = (TMOD & 0xf0) | 0x20;
#endif
	ser_set_spd(BDEFAULT);

#if defined(CONFIG_SERIAL_TIMER3)
#elif defined(CONFIG_USE_BRG)
#elif defined(CONFIG_SERIAL_TIMER2)
	TR2 = 1;
#else
	TR1 = 1;
#endif
#if CONFIG_I51_SERIAL_HIGHPRIO
	PS = 1;
#endif

	RI = 0;
	TI = 0;
#ifdef CONFIG_SERIAL_I51_STREAM
	i51_serial_interrupt_enable();
#endif
}

#if CONFIG_SERIAL_I51_STREAM && CONFIG_PORT
msg_ssize_t serial_i51_write(struct iocb __kernel *iocb) __reentrant
{
	msg_ssize_t ret = iocb->len;

	/* FIXME: don't wait in write() if MST_DONTWAIT is set */
	ser_tx_kernel(iocb->buf, ret);
	return ret;
}

msg_ssize_t serial_i51_read(struct iocb __kernel *iocb)
	__reentrant
{
	msg_ssize_t ret = iocb->len;

	ret = ser_rx_kernel(iocb->buf, ret);
	return ret;
}

unsigned char serial_i51_poll(port_t port)
	__reentrant
{
	unsigned char ret = 0;


	if (ser_rx_poll())
		ret |= POLLIN;
	return ret;
}
#endif

arch_initcall(i51_serial_init)
