/*
 * tmos - core.c
 *
 * spruked protocol stack
 *
 * Copyright (C) 2013..2014  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <tmos/spruked.h>
#include <tmos/crc7k.h>
#include <tmos/serial.h>
#include <tmos/watchdog.h>
#include <asm/softirq.h>

/*
 * initial checksum - we start from 0x7f to avoid CRC problems with
 * inserting zeros before checksum - in this application it's not needed
 * but it's a common practice.
 */
#define CHECKSUM_INIT		0x7f

#ifndef CRC7K_POLY
#define CRC7K_POLY	0x76
#endif

#define SPRUKED_SYNC		1

__bit spruked_softirq_pending = 1;

#if CONFIG_SPRUKED_STATS
static unsigned char spruked_rx_packets;
static unsigned char spruked_rx_crc_errors;

#define spruked_stats_rx_packet() \
	do { spruked_rx_packets++; } while (0)
#define spruked_stats_crc_error() \
	do { spruked_rx_crc_errors++; } while (0)
#else
#define spruked_stats_rx_packet() do { } while (0)
#define spruked_stats_crc_error() do { } while (0)
#endif

/*
 * The spruked stack manages four skb buffers, two I/O buffers for RX
 * and TX, and additional 3-entry forward queue.
 *
 * The I/O skbs are shared with spruked network driver.
 */
struct spruked_skb __kernel *spruked_rx_skb;
static struct spruked_skb __kernel *xmit_skb;
static struct spruked_skb __kernel *forward_skb;
static struct spruked_skb __kernel *f2_skb;
static struct spruked_skb __kernel *f3_skb;

void spruked_softirq(void) __using(SOFTIRQ_BANK)
{
	struct spruked_skb __kernel *skb;

retry:
	skb = spruked_rx_skb;
	/* process rx frame if needed */
	if (skb && !forward_skb) {
		u8 csum = SPRUKED_CHECKSUM_INIT;
		u8 __kernel *t;
		unsigned char i;

		spruked_rx_skb = NULL;

		smp_rmb();
		t = skb->payload;
		i = 4;
		do {
			csum = crc7k_byte(csum, *t);
			t++;
			i--;
		} while (i > 0);

		if (csum == SPRUKED_CHECKSUM_CHECK) {
			if (!f3_skb)
				f3_skb = skb;
			else if (!f2_skb)
				f2_skb = skb;
			else
				forward_skb = skb;
			spruked_stats_rx_packet();
		} else {
			cpu_flags_t flags;

			local_irq_save(flags);
			__spruked_kfree_skb(skb);
			local_irq_restore(flags);
			spruked_stats_crc_error();
		}
	}

	/*
	 * schedule frame to transmit if needed
	 */
	if (!spruked_tx_pending) {
		/*
		 * We are racing here with the serial interrupt handler.
		 * We must call spruked_xmit_start after reading
		 * spruked_tx_pending
		 */
		smp_rmb();
		if (f3_skb) {
			spruked_xmit_start(f3_skb);
			f3_skb = f2_skb;
			f2_skb = forward_skb;
			forward_skb = NULL;
			goto retry;
		} else if (xmit_skb) {
			spruked_xmit_start(xmit_skb);
			xmit_skb = NULL;
			set_need_resched();
		}
	}
}

void spruked_xmit(struct spruked_skb __kernel *skb)
{
	u8 __kernel *p = skb->payload;
	u8 csum = CHECKSUM_INIT;
	u8 hdr;
	u8 lo;
	u8 hi;

	hdr = *p++;
	hdr |= SPRUKED_SYNC;
	lo = *p++;
	hi = *p;
	hdr |= (((hi & 1) << 1) | (lo & 1)) << 1;
	lo &= 0xfe;
	hi &= 0xfe;

	p = skb->payload;
	*p = hdr;
	csum = crc7k_byte(csum, *p);
	p++;
	*p = lo;
	csum = crc7k_byte(csum, lo);
	p++;
	*p = hi;
	csum = crc7k_byte(csum, hi);
	p++;

	/*
	 * We use 7-bit CRC and we need to add single 0 bit here:
	 * multiple = (csum & 1) ? CRC_POLY : 0;
	 * csum = (csum >> 1) ^ multiple;
	 *
	 * Because we need to left shift CRC anyway we just do
	 * csum ^ (CRC_POLY << 1).
	 */
	if (csum & 1)
		csum ^= CRC7K_POLY << 1;

	*p = ~csum & 0xfe;

retry:
	local_irq_disable();
	if (xmit_skb) {
		safe_halt();
		goto retry;
	}

	xmit_skb = skb;
	spruked_softirq_pending = 1;
	local_irq_enable();

	call_softirq();
}

#if CONFIG_SPRUKED_STATS
char spruked_get_stats(struct spruked_stats __kernel *stats)
{
	stats->rx_packets = spruked_rx_packets;
	stats->rx_crc_errors = spruked_rx_crc_errors;
	return 0;
}
#endif
